from epstein_files.documents.documents.categories import Interesting, Neutral, Uninteresting
from epstein_files.documents.doj_files.full_text import EFTA00009622_TEXT
from epstein_files.documents.config.communication_cfg import CommunicationCfg, Platform, imessage_log, imessage_screenshot, skype_log, whatsapp_log
from epstein_files.documents.config.config_builder import letter
from epstein_files.documents.config.doc_cfg import NO_TRUNCATE, DocCfg
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.output.highlight_config import DITE_ANATA_JUILLIARD_URL
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import SVETLANA_NEWSGROUND
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.util.logging import logger

PRUSAKOVA_BERKELY = 'Epstein paid for Prusakova to go to Berkeley'
WOMEN_EMPOWERMENT = f"Women Empowerment (WE) conference"


GIRLS_CFGS = [
    DocCfg(id='019300', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} f. {KATHRYN_RUEMMLER}', date='2019-04-05', show_full_panel=True),
    DocCfg(id='022267', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} founder essay about growing the seminar business'),
    DocCfg(id='022407', author=SVETLANA_POZHIDAEVA, note=f'{WOMEN_EMPOWERMENT} seminar pitch deck'),
    EmailCfg(id='025874', note=f"possibly about {VIRGINIA_GIUFFRE} {QUESTION_MARKS}"),
    EmailCfg(id='030609', duplicate_ids=['030495']),

    # DOJ Files
    DocCfg(
        id='EFTA01334058',
        author='New York Post',
        category=Uninteresting.ARTICLE,
        date='2005-01-14',
        note="press snippets about Epstein's attempt to buy IMG Models as well as his relationships with Brett Ratner and Alina Pascau",
        is_in_chrono=False,
        is_interesting=True,
        truncate_to=1_200,
    ),
    DocCfg(
        id='EFTA01297974',
        note=f"DOJ queries OffshoreLeaks database for {SVETLANA_POZHIDAEVA}'s mother",
        show_full_panel=True,
        show_with_name=SVETLANA_POZHIDAEVA,
        truncate_to=500,
    ),
    DocCfg(id='EFTA02701609', author=GIANNI_SERAZZI, note='2morrow business plan', date='2014-09-30', date_uncertain=True),
    DocCfg(id='EFTA01078143', author=GIANNI_SERAZZI, note='2morrow modeling pitch deck', date='2014-09-30', date_uncertain=True),
    DocCfg(id='EFTA00306033', author=SERGEY_BELYAKOV, note='Epstein Russian visa', show_full_panel=True),
    DocCfg(
        id='EFTA01366289',
        author='Butterfly Trust',
        date='2015-03-30',
        is_interesting=20,
        note=f"transfer of $237,270 to {SVETLANA_POZHIDAEVA}'s father Raiffeisenbank from account at Standard Chartered Bank",
        show_full_panel=True,
        show_with_name=SVETLANA_POZHIDAEVA,
    ),
    DocCfg(id='EFTA00014648', note='former employee "understood that all the girls were school girls"', show_full_panel=True),
    DocCfg(id='EFTA00022133', note='table showing a list of Epstein accusations collection from public press reporting'),
    DocCfg(
        id='EFTA00077200',
        author=FRENCH_MINISTRY_OF_JUSTICE,
        date='2020-07-08',
        note=f'allegations that {JEAN_LUC_BRUNEL} drugged a model and Epstein + Ghislaine assaulted a minor',
    ),
    DocCfg(
        id='EFTA00079597',
        author=FRENCH_MINISTRY_OF_JUSTICE,
        note=f"allegations of sexual assault by {JEAN_LUC_BRUNEL}",
        date='2020-07-08',
    ),

    # Communications
    CommunicationCfg(
        id='EFTA01612665',
        note='Epstein gives advice on how to recruit girls',
        highlight_quote="stressed about finding girls. It's hard",
    ),
    imessage_log(
        id='EFTA02731576',
        is_interesting=10,
        note=f"making contemporaneous accusations",
        recipients=[LEON_BLACK],
        show_full_panel=True,
    ),
    imessage_screenshot('EFTA02731525', author=LEON_BLACK, author_uncertain=True),
    imessage_screenshot('EFTA01618381', author=RENATA_BOLOTOVA, author_uncertain='sneaky', highlight_quote='how many new girls are there in the island W you now?'),
    imessage_screenshot('EFTA01622387', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    imessage_screenshot('EFTA01618494', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    imessage_screenshot('EFTA01618400', author=RENATA_BOLOTOVA, author_uncertain='sneaky'),
    letter('EFTA00078198', 'Marsh Law Firm', [USANYS], "allegations of sexual abuse at Epstein's house", '2020-11-10'),
    skype_log(
        id='EFTA01616933',
        date='2019-04-08',
        note='"no one can beat your pussy network"',
        highlight_quote='now im finding pussy for you',
    ),
    skype_log(
        id='EFTA01623342',
        author=ANNA_KASATKINA,
        author_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        note='recruiting russian girls',
        url='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
    ),
    whatsapp_log(
        id='EFTA01612733',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        highlight_quote='way better than supplying you ladies',
        note=f'saying being Epstein\'s lawyer would be preferable to recruiting, discusses Medici Bank',
    ),

    # Authors
    EmailCfg(
        id='EFTA02339926',
        author=JULIA_SANTOS,
        author_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        highlight_quote='very sweet and might be naughty t=o',
    ),
    EmailCfg(
        id='EFTA01745650',
        author=KIRA_DIKHTYAR,
        author_uncertain='AOL signature',
        note=f'{SANITA} is big mad at Epstein and "makes lists" which is freaking everyone out',
        show_with_name=SANITA,
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA01781620', author=KIRA_DIKHTYAR, author_uncertain='"Sent from AOL Mobile Mail" in chain'),
    EmailCfg(id='EFTA01805304', author=KIRA_DIKHTYAR, author_reason='Jmail', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01766199', author=KIRA_DIKHTYAR, author_reason='reply', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='033178',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        is_interesting=10,
        note=f'Masha Prusso asks about {ZUBAIR_KHAN}, discusses recruiting girls for Epstein',
    ),
    EmailCfg(
        id='EFTA00669538',
        author=MARIA_PRUSAKOVA,
        author_uncertain='discussion of "experiments"',
        note='blowjob instructions, sexual experiments',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA02010589',
        author=MARIA_PRUSAKOVA,
        date='2012-01-31 9:16 PM',
        note=f'raunchy email mentioning "Sasha Grey guy", note {BORIS_NIKOLIC} and bgC3 in SMTP info at end',
        is_interesting=10,
        recipients=[JEFFREY_EPSTEIN],
    ),
    EmailCfg(
        id='EFTA00928879',
        author=MARIA_PRUSAKOVA,
        author_uncertain='sounds like her',
        note=f'raunchy email',
        highlight_quote='your job is to find interesting girls',
        truncate_to=3_300,
    ),
    EmailCfg(
        id='EFTA01841553',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        highlight_quote='being fucked in the ass hurts me a lot',
        truncate_to=900,
    ),
    EmailCfg(
        id='032374',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        note='boyfriend named Christian mentioned',
        is_interesting=True,
    ),
    EmailCfg(id='032375', author=MARIA_PRUSAKOVA, author_reason=PRUSAKOVA_BERKELY, truncate_to=295),
    EmailCfg(id='033246', author=MARIA_PRUSAKOVA, author_reason='Kind regards/Cordialement in signature'),
    EmailCfg(id='EFTA01990879', author=MARIA_PRUSAKOVA, author_uncertain='Clifford Chance, could be Miranda'),
    EmailCfg(id='EFTA01990168', author=MARIA_PRUSAKOVA, author_uncertain='"experiments"', is_interesting=True),
    EmailCfg(id='EFTA00671662', author='Miranda', author_reason='quoted signature', note="yet another girl finder"),
    EmailCfg(id='EFTA01047249', author=NADIA_MARCINKO, author_reason='Miro, https://archive.ph/Qa6vU#selection-1621.160-1621.184'),
    EmailCfg(id='EFTA02441035', author=STEVEN_VICTOR_MD, note="complaints about free medical treatment for Epstein's girls"),
    EmailCfg(
        id='EFTA00659941',
        author=SVETLANA_POZHIDAEVA,
        author_reason='https://tommycarstensen.com/epstein/findings.html#victim-diary-names-multiple-men',
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA} forwarding her intimate conversations with {JOSHUA_FINK} to Epstein (to what end?)',
    ),
    EmailCfg(
        id='EFTA01844498',
        author=SVETLANA_POZHIDAEVA,
        author_reason='Josh Fink, BlackBerry signature',
        is_interesting=10,
        note=f"jealous boyfriend {JOSHUA_FINK} accuses Svetlana had an affair with Navalny {QUESTION_MARKS}",
    ),
    EmailCfg(
        id='EFTA01870453',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain=f"{JOSHUA_FINK} texts",
        highlight_quote="You have known you are preg for a week",
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA} fwds a discussion about an abortion to Epstein',
    ),
    EmailCfg(
        id='EFTA01894879',
        note='"most of all its cheap excort"',
        duplicate_ids=['EFTA00876711'],
        highlight_quote='400 girls for model agencies and wedding agencies',
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain='"Svet"',
    ),
    EmailCfg(
        id='EFTA01991293',
        author=SVETLANA_POZHIDAEVA,
        author_reason=f"convo about {JOSHUA_FINK} breakup",
        is_interesting=10,
        note=f'did {ELON_MUSK} date {SVETLANA_POZHIDAEVA} after she blackmailed {JOSHUA_FINK}? ("he asked 6 times if Elon [gave] me anything")',
    ),
    EmailCfg(
        id='EFTA01998247',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='device signature',
        note=f'trying to recruit {ELON_MUSK}?',
        is_interesting=10,
        show_with_name=ELON_MUSK,
    ),
    EmailCfg(
        id='EFTA01894132',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='"Svet" in unredacted reply',
        truncate_to=(900, 1_600),
    ),
    EmailCfg(id='EFTA01024246', note=f"invoice from {SVETLANA_POZHIDAEVA}'s WE Talks", show_with_name=SVETLANA_POZHIDAEVA),
    EmailCfg(id='EFTA02634573', author=SVETLANA_POZHIDAEVA, note='updating Epstein on Women Empowerment'),
    EmailCfg(id='EFTA01772677', author=SVETLANA_POZHIDAEVA, author_uncertain='device signature'),
    EmailCfg(id='EFTA01877224', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts", is_interesting=10),
    EmailCfg(id='EFTA00937507', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts", is_interesting=10),
    EmailCfg(
        id='EFTA01744405',
        author=SVETLANA_POZHIDAEVA,
        is_interesting=10,
        note=f"{SVETLANA_POZHIDAEVA} discusses her father (an FSB officer) and 'Josh' (presumably {JOSHUA_FINK})",
        url=SVETLANA_NEWSGROUND,
    ),
    EmailCfg(
        id='EFTA02475964',
        note=f"Epstein thanks {SVETLANA_POZHIDAEVA}'s father, a former (?) FSB officer",
        is_interesting=20,
        recipients=[SVETLANA_POZHIDAEVA],
        url=SVETLANA_NEWSGROUND,
    ),
    EmailCfg(
        id='EFTA02480401',
        highlight_quote='i re=iewed payment of 237,270 for svet family',
        is_interesting=20,
        note=f'"svet family" is {SVETLANA_POZHIDAEVA}\'s family in Russia',
        show_with_name=SVETLANA_POZHIDAEVA,
    ),
    EmailCfg(id='EFTA02431535', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA01875607', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA00901581', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA02449477', author=UNKNOWN_GIRL, note='argument about finding girls for Epstein', is_interesting=10, truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02027844', author=UNKNOWN_GIRL, note='permission to "fuck someone with a condom"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00763537', author=UNKNOWN_GIRL, highlight_quote="19 but I look younger than her"),
    EmailCfg(id='EFTA01782788', author=UNKNOWN_GIRL, highlight_quote='I need 21-24, wiling to travel and work hard', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='EFTA01024118',
        author=YULIA_DOROKHINA,
        author_reason='juliador89@mail.ru visible in verticle text',
        highlight_quote='I sent you nude pictures, sexy videos even masturbation',
        truncate_to=1_900,
    ),
    EmailCfg(
        id='EFTA01035876',
        author=YULIA_DOROKHINA,
        author_uncertain='"Yulia" in reply',
        note=f'looking for a job under {JIDE_ZEITLIN}',
        show_with_name=JIDE_ZEITLIN,
    ),
    EmailCfg(
        id='EFTA01036212',
        author=YULIA_DOROKHINA,
        author_uncertain='"Yulia"',
        note=f"Epstein wants {YULIA_DOROKHINA} to travel on short notice",
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA01025400', author=YULIA_DOROKHINA),
    EmailCfg(id='EFTA01025734', author=YULIA_DOROKHINA, truncate_to=1_300),
    EmailCfg(id='EFTA01024680', author=YULIA_DOROKHINA, author_uncertain='thread'),
    EmailCfg(id='EFTA01024729', author=YULIA_DOROKHINA, author_uncertain='thread'),
    EmailCfg(id='EFTA02634523', author=YULIA_DOROKHINA, author_uncertain='"Yulia"'),
    EmailCfg(id='EFTA01028241', author=YULIA_DOROKHINA, author_uncertain='"Yulia" in other emails in thread', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01026136', author=YULIA_DOROKHINA, highlight_quote='send the masturbation video', truncate_to=5000),

    # Recipients
    EmailCfg(
        id='EFTA02223525',
        recipients=[ANNA_KASATKINA, BELLA_KLEIN, JANUSZ_BANASIAK],
        recipient_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        truncate_to=800
    ),
    EmailCfg(id='EFTA02470883', recipients=[YULIA_DOROKHINA]),
    EmailCfg(
        id='EFTA00997627',
        highlight_quote='you will be introduced to more gorgeous Alicas, i promise',
        recipients=[MARIA_PRUSAKOVA],
        recipient_uncertain='Kind regards/Cordialement in signature',
    ),
    EmailCfg(id='EFTA00719146', recipients=[MARIA_PRUSAKOVA], recipient_uncertain='Clifford Chance, could be Miranda'),
    EmailCfg(
        id='EFTA01857628',
        note='"igrushki" means "toys" in Russian',
        recipients=[RENATA_BOLOTOVA],
        recipient_uncertain=True,
        url='https://x.com/FlippersUpNow/status/2023457206109155345',
    ),
    EmailCfg(
        id='EFTA00857669',
        note=f"{SANITA} is maybe the only great girl with right thinking in the entire Epstein files",
        recipients=[SANITA],
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA00738485', recipients=[STEVEN_VICTOR_MD], truncate_to=640),
    EmailCfg(
        id='EFTA02560884',
        is_interesting=10,
        note=f"is Epstein drafting {SVETLANA_POZHIDAEVA}'s romantic responses to {JOSHUA_FINK} {QUESTION_MARKS}",
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain=f"{JOSHUA_FINK} texts",
    ),
    EmailCfg(
        id='EFTA02543909',
        author_reason='p.selena@yahoo.com in replies',
        is_interesting=10,
        note=f'{SVETLANA_POZHIDAEVA}\'s relationship with "J" ({JOSHUA_FINK}?) and "B" {QUESTION_MARKS}',
        recipients=[SVETLANA_POZHIDAEVA],
    ),
    EmailCfg(
        id='EFTA00937556',
        is_interesting=10,
        note=f"looks like Epstein coaching {SVETLANA_POZHIDAEVA} on how to respond to {JOSHUA_FINK}",
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain=True,
    ),
    EmailCfg(
        id='EFTA02447019',
        url='https://www.wsj.com/us-news/epstein-victim-svetlana-pozhidaeva-22f2ebf6',
        recipients=[SVETLANA_POZHIDAEVA],
    ),
    EmailCfg(
        id='EFTA00888467',
        is_interesting=True,
        note=f'very strange conversation about "Peter" (maybe {PETER_MANDELSON}?) and "mj" (maybe {HASSAN_JAMEEL}?)',
        recipients=[UNKNOWN_GIRL],
    ),
    EmailCfg(id='EFTA00897668', recipients=[UNKNOWN_GIRL], entity_names=[JEAN_LUC_BRUNEL, JEFFREY_EPSTEIN, UNKNOWN_GIRL]),
    EmailCfg(id='EFTA00848644', recipients=[UNKNOWN_GIRL], highlight_quote='take a selfie of your pussy and send'),
    EmailCfg(
        id='EFTA01026268',
        author_reason='juliador89®mail.ru is unredacted',
        is_interesting=10,
        note=f"Epstein pressuring {YULIA_DOROKHINA} to find girls in Russia and send him nude pics and videos",
        recipients=[YULIA_DOROKHINA],
        truncate_to=3_100,
    ),
    EmailCfg(id='EFTA01024839', recipients=[YULIA_DOROKHINA], recipient_uncertain='part of thread'),
    EmailCfg(id='EFTA02276109', recipients=[YULIA_DOROKHINA], recipient_uncertain='"Yulia"'),

    # Descriptions
    EmailCfg(id='EFTA00078209', note='allegations about Russian organized crime, sex trafficking', truncate_to=8_500),
    EmailCfg(id='EFTA01648951', note=f'allegations against Trump, {HOWARD_LUTNICK}, Glen Dubin, {LEON_BLACK}, {JES_STALEY}'),
    EmailCfg(id='EFTA00950368', note='"any [girl] friends for me?"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01814983', note=f"Epstein and {NICOLE_JUNKERMANN} used to date", show_with_name=NICOLE_JUNKERMANN),
    EmailCfg(id='EFTA02049364', note='Epstein arranging for healthcare for five girls', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02209622', note=f'flight to Riga for {LASMA_KUHTARSKA}', show_with_name=LASMA_KUHTARSKA, truncate_to=600),
    EmailCfg(id='EFTA00968086', note='request for help for a Juilliard student', url=DITE_ANATA_JUILLIARD_URL),
    EmailCfg(
        id='EFTA00883738',
        note=f'Epstein recruiter {DASHA_GRUPMAN} looking to spend $50,000 to rent a townhouse for "diplomats"',
        show_with_name=DASHA_GRUPMAN,
    ),
    DocCfg(
        id='EFTA00009622',
        date='2006-07-19',
        note='handwritten notes from a victim interview transcribed by Claude AI',
        is_interesting=True,
        display_text=EFTA00009622_TEXT,
    ),
    EmailCfg(
        id='EFTA00577409',
        date='2003-06-01',
        date_uncertain='just a guess, string is 4501 which is NULL in Microsoft Outlook',
        note='Epstein and Ghislaine discuss girl "from donad trumps party" who blackmailed [someone]',
    ),
    EmailCfg(
        id='EFTA02007083',
        note=f"{BORIS_NIKOLIC} encourages {KIMBAL_MUSK} to ditch his girlfriend for Epstein's girl {JENNIFER_KALIN} (and he says yes)",
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA00412260',
        author=LESLEY_GROFF,
        comment='very broken file',
        date='2012-05-25 19:51:20',
        note=f'emphasis quotes by {LESLEY_GROFF}',
        highlight_quote='Jeffrey was asking if you and "your friend" could come by and meet Peter',
        recipients=[None],
        truncate_to=(6_700, 7_450),
    ),
    EmailCfg(
        id='EFTA00438431',
        note=f"visa invitation for {KARYNA_SHULIAK} to visit {QUESTION_MARKS}",
        recipients=[LESLEY_GROFF],
        show_with_name=KARYNA_SHULIAK,
    ),

    # Quotes
    EmailCfg(
        id='EFTA02504370',
        highlight_quote=f'lovely girl with financial troubles',
        recipients=[KIRA_DIKHTYAR],
        recipient_uncertain='"Sent from AOL Mobile Mail"',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA01022353',
        highlight_quote='she said she was 14-15 yo',
        note='(George Models is basically a porn site)',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01885139',
        highlight_quote='Kimbal really like Jen',
        is_interesting=True,
        note=f'({JENNIFER_KALIN}, Epstein\'s girl whom {KIMBAL_MUSK} soon dated)',
    ),
    EmailCfg(
        id='EFTA02415420',
        highlight_quote="I have a friend of Putin,s",
        is_interesting=10,
    ),
    EmailCfg(
        id='EFTA02555750',
        highlight_quote='suspend any billing from now on until further notification from you or the model',
        note=f"who's \"the model\" {JEAN_LUC_BRUNEL} is billing {JOSHUA_FINK} for?",
        is_interesting=10,
    ),
    EmailCfg(id='EFTA00971793', highlight_quote='Attached is resume of Lutnick nanny', note=f"why does Epstein need a nanny?"),
    EmailCfg(id='EFTA01965732', highlight_quote='facilictating his illicit trysts, with married women, to being asked to provide adderall fro bridge tournamnts'),
    EmailCfg(id='EFTA01768670', highlight_quote='find girls for the agency', note=f'but {BORIS_NIKOLIC} is a "biotech investor"...'),
    EmailCfg(id='EFTA01905320', highlight_quote='girls and i are going to see elon musk at space x tomorrow', is_interesting=True),
    EmailCfg(id='EFTA01803353', highlight_quote='having problems with regina', note='(Ramsey?)'),
    EmailCfg(id='EFTA01858707', highlight_quote='How is the marriage plan going'),
    EmailCfg(id='EFTA00950394', highlight_quote='how old will you go?'),
    EmailCfg(id='EFTA02499884', highlight_quote='I gave another girl to kimball and he is thrilled', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02557757', highlight_quote="I have a girl Sana working half time/who by the way you'd like"),
    EmailCfg(id='EFTA00771856', highlight_quote="I know 23 is on the old side for you", is_interesting=True),
    EmailCfg(id='EFTA01974447', highlight_quote='I know you are going to meet putin on the 20th', is_interesting=True),
    EmailCfg(id='EFTA01767237', highlight_quote='is 24 too old for you?', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00854166', highlight_quote='kazak contract will be ready for your review', is_interesting=True, truncate_to=1100),
    EmailCfg(
        id='EFTA00878421',
        highlight_quote="said that she felt gods presence next to her when she was in bed",
        is_interesting=True,
        recipients=[NADIA_MARCINKO],
        recipient_uncertain='https://archive.ph/Qa6vU',  # https://www.nybooks.com/articles/2026/03/26/the-devil-himself-jeffrey-epstein-enright/
    ),
    EmailCfg(id='EFTA00894079', highlight_quote='say hi to Snow White'),
    EmailCfg(id='EFTA00775255', highlight_quote='she is 20 years old but she looks younger', non_participants=[BROCK_PIERCE]),
    EmailCfg(id='EFTA00366629', highlight_quote='stunning girls to bring over drinks when your next available'),
    EmailCfg(id='EFTA02387676', highlight_quote="thats the closest i get to karyna providing me new pussy"),
    EmailCfg(id='EFTA00908180', highlight_quote='two cinderellas'),
    EmailCfg(id='EFTA01930285', highlight_quote='ukraine upheaval should provide many opportunites'),
    EmailCfg(id='EFTA00751119', highlight_quote='Valdson to teach girls how to serve'),
    EmailCfg(id='EFTA01995972', highlight_quote='you are meeting elon tmr for lunch'),
    EmailCfg(id='EFTA01998027', highlight_quote="What day/night will be the wildest party on your island?", is_interesting=True),
    EmailCfg(id='EFTA01930501', highlight_quote="Your littlest girl was a little naughty"),
    EmailCfg(id='EFTA01926961', author=SVETLANA_POZHIDAEVA, highlight_quote="If you take too many more trips to Russia, I'm going to assume that you're FSB/ KGB by your parents"),
    EmailCfg(id='DropSite 2006-05-29 1329', highlight_quote='I am giving the little girl a modeling\n> lesson'),

    # Josh Fink
    EmailCfg(
        id='EFTA01868066',
        # author=SVETLANA_POZHIDAEVA,
        # author_uncertain=f"{JOSHUA_FINK} + Sent from Blackberry",
        note=f'impersonating [someone] talking to {JOSHUA_FINK}',
        duplicate_ids=['EFTA01995523'],
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01877084',
        comment='https://x.com/HansMahncke/status/2024351830998994979',
        note=f'Epstein and {JEAN_LUC_BRUNEL} "managing" {JOSHUA_FINK}\'s relationship with {SVETLANA_POZHIDAEVA}',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00704924',
        note=f"stalking Josh Fink (attachment is pic of {JOSHUA_FINK} and {FILIPA_PEROVIC} who later married)",
        show_with_name=JOSHUA_FINK,
    ),
    EmailCfg(
        id='EFTA02217672',
        note=f"{DAVID_L_NEUHAUSER} is raising money for a {JOSHUA_FINK} fund, promises Epstein he won't reveal they're talking",
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01907281',
        note=f"Epstein asks about {JOSHUA_FINK}, pretends it's because {TANCREDI_MARCHIOLO} asked *him* (Epstein asked Marchiolo)",
        show_with_name=JOSHUA_FINK,
        is_interesting=True,
    ),

    # Descriptions
    EmailCfg(
        id='EFTA02627361',
        note=f'Epstein sending {YULIA_DOROKHINA} {QUESTION_MARKS} to {JIDE_ZEITLIN}',
        recipients=[JIDE_ZEITLIN],
        recipient_uncertain=True,
        show_with_name=YULIA_DOROKHINA,
    ),
    EmailCfg(
        id='EFTA02434682',
        note=f"Epstein and {JEAN_LUC_BRUNEL} courting a finance bro with women",
        is_interesting=True,
        truncate_to=1500,
    ),
    EmailCfg(id='EFTA01849371', note=f"{SVETLANA_POZHIDAEVA} meets Larry Fink", truncate_to=500),
    # EmailCfg(id='EFTA01849371', note=f"{SVETLANA_POZHIDAEVA} meets Nat Rothschild", show_with_name=SVETLANA_POZHIDAEVA),
    EmailCfg(id='EFTA02605514', author=SVETLANA_POZHIDAEVA, note=f"{SVETLANA_POZHIDAEVA} meets {BILL_GATES}"),
    EmailCfg(id='EFTA01895036', note=f"Epstein invites {ELON_MUSK} to Paris", highlight_quote='svetlana will be in paris 17-20'),
    EmailCfg(id='EFTA02142175', note=f"{DARREN_INDYKE} meeting {JEAN_LUC_BRUNEL} at Karin Models office"),
    EmailCfg(id='EFTA01840103', note="Epstein apparently suggesting amphetamines (adderall)", is_interesting=10),
    EmailCfg(id='EFTA01987855', note=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} very interested in a Liberian sex scandal'),
    EmailCfg(id='EFTA01817903', note=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} sharing news of a Liberian sex scandal', is_interesting=10),
    EmailCfg(id='EFTA01818540', note=f'Epstein forwarding news of the Liberian sex scandal to {JES_STALEY}'),
    EmailCfg(id='EFTA01140210', note='Epstein asks about fake Instagram followers'),
    EmailCfg(id='EFTA00937981', note=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA00668344', note=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA01856467', note=f"Epstein looking for {JOSHUA_FINK} (again)", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01854125', note=f'Epstein says to {HENRY_JARECKI} "you torture, and mistreat each" girl'),
    EmailCfg(id='EFTA01936438', note=f'Epstein\'s standard password might be "MVEMJSNUP"', is_interesting=10),
    EmailCfg(id='EFTA02559808', note=f"{EVA_DUBIN} delivering amphetamines (adderall) to Epstein", is_interesting=10, truncate_to=200),
    EmailCfg(id='EFTA01953412', note=f'{EVA_DUBIN} suggests Epstein endow "The Epstein Floor For Women" at Mt. Sinai'),
    EmailCfg(id='EFTA00658028', note=f'{FAITH_KATES} trying to get a modeling contract for "Regina"?'),
    EmailCfg(id='EFTA00855566', note=f"introducing {SANITA}", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01145923', note=f"{JEAN_LUC_BRUNEL} modeling contract for {MC2_MODEL_MGMT}"),
    EmailCfg(id='EFTA01660651', note='list of Trump accusers', is_interesting=True),
    EmailCfg(id='EFTA02609062', note=f"{MASHA_DROKOVA} is assembling a team", is_interesting=True),
    EmailCfg(id='EFTA02557291', note='possibly recruiting girls'),
    EmailCfg(
        id='EFTA01811230',
        note=f'"Renat" is probably {RENATA_BOLOTOVA}, {REDACTED} might be "Irina" or {SVETLANA_POZHIDAEVA}',
        duplicate_ids=['EFTA00732294'],
    ),
    EmailCfg(id='EFTA00927927', note=f"{RENATA_BOLOTOVA} possibly talking about recruiting girls"),
    EmailCfg(id='EFTA01816788', note=f"Russian girl for {PRINCE_ANDREW}"),
    EmailCfg(id='EFTA00876739', note=f'scouting girls in Ukraine {QUESTION_MARKS}'),
    EmailCfg(id='EFTA00901905', note="smoking in the house oh no!"),
    EmailCfg(id='EFTA00914505', note=f'{SVETLANA_POZHIDAEVA} and {JOSHUA_FINK}...', duplicate_ids=['EFTA01999588']),
    EmailCfg(id='EFTA01798715', note=f"taking an interest in {JOSHUA_FINK}'s investments in Eritrea in between hair gigs"),
    EmailCfg(id='EFTA00572157', note='translation: "Congratulations on the new addition!!! What did you name the little miracle?" (seems to be about a cat)'),
    EmailCfg(id='EFTA00631762', note="visa problems for Epstein's South African friend"),
    EmailCfg(id='EFTA00376832', show_with_name=MARIANA_IDZKOWSKA),
    EmailCfg(id='EFTA00743526', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02664956', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00689698', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02702141', truncate_to=NO_TRUNCATE, comment=ELON_MUSK),
    EmailCfg(id='EFTA00819508', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00460312', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00878255', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02551185', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02538269', comment=BORIS_NIKOLIC, duplicate_ids=['EFTA01866636']),
    EmailCfg(id='EFTA00736518', comment=JES_STALEY + ' beauty and the beast'),
    EmailCfg(id='EFTA02344845', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA02491007', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00633284', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00664521', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00997666', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00964428', comment=JULIA_SANTOS + ' Valdson'),
    EmailCfg(id='EFTA02398135', comment=JULIA_SANTOS + ' "Sochi"'),
    EmailCfg(id='EFTA00766770'),
]
